/*
 * common_functions.cpp
 *
 *  Created on: Sep 22, 2015
 *      Author: abhattac
 */

#include <iostream>
#include <algorithm>
#include <unordered_map>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <cstdint>

#include "../misc/Pattern.h"
#include "../misc/ChoicePattern.h"
#include "../misc/EncodingCalucator.h"
#include <math.h>

#include "../misc/Data.h"
#include "../misc/DataCover.h"
#include "../misc/Mwindow.h"
#include "../misc/LabelList.h"

using namespace std;

static unordered_map<uint32_t, double> log_gma;
static unordered_map<uint32_t, double> log_dub_fact;

bool vec_and( vector<bool> &v )
{
	bool a = true;
	for( uint32_t i = 0; i < v.size(); i++ )
	{
		a = a & v[i];
	}
	return a;
}

void print_all_patterns( vector<Pattern*> P )
{
	cout << "Printing all patterns: " << endl;
	double avg_length = 0;
	for(unsigned int i = 0; i < P.size(); i++ )
	{
		cout << "Pattern: " << i << endl;
		for(int j = 0; j < P[i]->getSize(); j++ )
		{
			cout << P[i]->charAt(j) << " - ";
		}
		avg_length += double( P[i]->getSize() );
		cout << " Usage: " << P[i]->getUsage() << endl;
	}
	cout << "Average pattern length: " << avg_length / P.size() << endl;
}

void delete_list(Mwindow *start)
{
	Mwindow *itr = start, *itr_prev = nullptr;
	while( itr != nullptr )
	{
		itr_prev = itr;
		itr = itr->next;
		delete itr_prev;
	}
}

double get_log_gamma( int key, EncodingCalucator &ec )
{
	double lg_gma;
	double &stored_val = log_gma[ key ];
	if (stored_val)
	{
		lg_gma = stored_val;
	}
	else
	{
		lg_gma = ec.log_gamma( key );
		log_gma[ key ] = lg_gma;
	}
	return lg_gma;
}

double get_log_dub_fact( int key, EncodingCalucator &ec )
{
	double lg_dub_fact;
	double &stored_val = log_dub_fact[ key ];
	if (stored_val)
	{
		lg_dub_fact = stored_val;
	}
	else
	{
		lg_dub_fact = ec.ln_double_factorial_exact( key );
		log_dub_fact[ key ] = lg_dub_fact;
	}
	return lg_dub_fact;
}

double approx_prequential_length_of_binary_sequence( uint32_t f1, uint32_t f2 )
{
	if( f1 == 0 || f2 == 0 )
	{
		return 0;
	}
	EncodingCalucator ec;
	int f = f1 + f2;
	double lg_gma1, lg_gma2, lg_dub_fact1, lg_dub_fact2;

	lg_gma1 = get_log_gamma( (f + 1), ec );
	lg_gma2 = get_log_gamma( 1, ec );
	lg_dub_fact1 = get_log_dub_fact( (2 * f1 - 1), ec );
	lg_dub_fact2 = get_log_dub_fact( (2 * f2 - 1), ec );


	return lg_gma1  - lg_gma2
			- ( lg_dub_fact1  - f1 )
			- ( lg_dub_fact2  - f2 );
}

double optimal_length_of_binary_sequence( uint32_t f1, uint32_t f2  )
{
	if( f1 > 0 && f2 >0 )
	{
		return f1 * log2( ((double) (f1 + f2)) / f1 ) + f2 * log2( ((double) (f1 + f2)) / f2 );
	}
	else
	{
		return 0;
	}
}

double exact_prequential_length_of_binary_sequence( int fills, int gaps, double epsilon )
{
	double preq_code_len = 0;
	EncodingCalucator ec;
	for(int i = 0; i < fills; i++)
	{
		preq_code_len += ec.lu( (double) (2*epsilon + i) /  (epsilon + i) );
	}
	for(int i = 0; i < gaps; i++)
	{
		preq_code_len += ec.lu( (double) (2*epsilon + fills + i) /  (epsilon + i) );
	}
	return preq_code_len;
}

double exact_prequential_length_of_trinary_sequence( int fills, int gaps, int stops, double epsilon )
{
	double preq_code_len = 0;
	EncodingCalucator ec;
	for(int i = 0; i < fills; i++)
	{
		preq_code_len += ec.lu( (double) (3*epsilon + i) /  (epsilon + i) );
	}
	for(int i = 0; i < gaps; i++)
	{
		preq_code_len += ec.lu( (double) (3*epsilon + fills + i) /  (epsilon + i) );
	}
	for(int i = 0; i < stops; i++)
	{
		preq_code_len += ec.lu( (double) (3*epsilon + fills + gaps + i) /  (epsilon + i) );
	}
	return preq_code_len;
}

double gain(Mwindow *w, Data *D)
{

	const Pattern *p = w->pattern;
	double lx = 0;
	for(int i=0; i<w->pattern->getSize(); i++)
	{
		lx += log2((double)D->getSize()/D->getAlphabetSupport( p->charAt(i) ));
	}
	/*if( w->pattern->getGapCodeLength() != w->pattern->getGapCodeLength() )
	{
		cout << "p code len is less than zero: " << w->pattern->getPatternCodeLength() << endl;
		exit(0);
	}
	if( w->getNumOfGaps() > w->pattern->getSize() )
	{
		return (-(w->pattern->getPatternCodeLength())
				- (w->getEndIndex() - w->getStartIndex() - w->pattern->getSize() + 1)*1
				-(w->pattern->getSize() - 1)*1 + lx);
		//cout <<  w->gaps << " - " << (-(w->X->getPatternCodeLength()) - (w->j - w->i - w->X->getSize() + 1)*(w->X->getGapCodeLength()) -(w->X->getSize() - 1)*(w->X->getFillCodeLength()) + lx) << endl;
	}*/

	double gain = -(w->pattern->getPatternCodeLength())
					- (w->getEndIndex() - w->getStartIndex() - w->pattern->getSize() + 1) * (w->pattern->getGapCodeLength())
					-(w->pattern->getSize() - 1)*(w->pattern->getFillCodeLength()) + lx;

	/*cout << " Mwindow gaps: " << w->getNumOfGaps()
		 << " pattern number: " << w->pattern->getPatternNum()
		 << " pattern usage: " << w->pattern->getUsage()
		 << " pattern gaps: " << w->pattern->getGaps()
		 << " fill code length: " << w->pattern->getFillCodeLength()
		 << " gap code length: " << w->pattern->getGapCodeLength()
		 << " gain: " << gain
		 << endl;*/

	return gain;
}

double gain_preq(Mwindow *w, Data *D)
{

	const Pattern *p = w->pattern;
	double lx = 0;
	for(uint32_t i = 0; i < w->pattern->getSize(); i++)
	{
		lx += log2((double)D->getSize()/D->getAlphabetSupport( p->charAt(i) ));
	}

	return (-(w->pattern->getPatternCodeLength())
			- (w->getEndIndex() - w->getStartIndex() - w->pattern->getSize() + 1) * (w->pattern->getPreqGapCodeLength())
			-(w->pattern->getSize() - 1)*(w->pattern->getPreqFillCodeLength()) + lx);
}



double new_gain(Mwindow *w, Data *D)
{
	EncodingCalucator ec;
	//cout << (w->i) << " ";
	const Pattern *p = w->pattern;
	double lx = 0;
	for(uint32_t i = 0; i < p->getSize(); i++)
	{
		lx += log2((double)D->getSize()/D->getAlphabetSupport( p->charAt(i) ));
	}
	return (-(w->pattern->getPatternCodeLength()) - w->pattern->getGapCodeLength( w->getNumOfGaps() )  + lx);
}

int length_list(Mwindow *allwin_start)
{
	int N = 0;
	Mwindow *allwin_tmp = allwin_start;
	while( allwin_tmp != nullptr )
	{
		N++;
		allwin_tmp = allwin_tmp->next;
	}
	return N;
}

/*void print_all_mwindow_patterns(Mwindow *start)
{
	Mwindow *tmp = start;
	while( tmp != nullptr )
	{
		cout << "Pos: " << tmp->getStartIndex() << " to " << tmp->getEndIndex() << " | ";
		tmp->pattern->printPattern();
		tmp = tmp->next;
	}
	exit(0);
}*/

int max_pattern_num(vector<Pattern*> &list_of_patterns)
{
	uint32_t max = 0;
	for(uint32_t i = 0; i < list_of_patterns.size(); i++)
	{
		if( list_of_patterns[i]->getPatternNum() > max )
		{
			max = list_of_patterns[i]->getPatternNum();
		}
	}
	return max;
}

bool check_if_pattern_num_exists_twice(vector<Pattern*> &list_of_patterns, uint32_t pat_num)
{
	uint32_t count = 0;
	for(uint32_t i = 0; i < list_of_patterns.size(); i++)
	{
		if( list_of_patterns[i]->getPatternNum() == pat_num )
		{
			count++;
		}
		if( count == 2 )
		{
			return true;
		}
	}
	return false;
}

void pat_num_check( vector<Pattern*> &list_of_patterns )
{
	for(uint32_t i = 0; i < list_of_patterns.size(); i++)
	{
		if( check_if_pattern_num_exists_twice( list_of_patterns, list_of_patterns[i]->getPatternNum() ) )
		{
			list_of_patterns[i]->setPatternNum( max_pattern_num( list_of_patterns ) + 1 );
		}
	}

}

void delete_mwindows_in_vector( vector<Mwindow*> &mwindow_list )
{
	for_each( mwindow_list.begin(), mwindow_list.end(), [](Mwindow *w){ delete w; });
}

uint32_t max(uint32_t a, uint32_t b)
{
	if( a > b )
		return a;
	else
		return b;
}



void print_all_choice_patterns( vector<ChoicePattern> &list_of_choice_patterns )
{
	cout << " ------------ All choice patterns -------- " << endl;
	for(uint32_t i = 0; i < list_of_choice_patterns.size(); i++)
	{
		list_of_choice_patterns[i].print();
	}
	cout << " ----------------------------------------- " << endl;
}

void print_all_choice_patterns( vector<ChoicePattern> &list_of_choice_patterns, Data *D )
{
	cout << " ------------ All choice patterns -------- " << endl;
	for(uint32_t i = 0; i < list_of_choice_patterns.size(); i++)
	{
		list_of_choice_patterns[i].print();
		cout << "codelen: " << list_of_choice_patterns[i].get_codelen(D) << endl;
	}
	cout << " ----------------------------------------- " << endl;
}

int32_t check_if_new_pattern_can_be_made_a_choice_pattern( Pattern *new_pattern, vector<ChoicePattern> &list_of_choice_patterns, Data *D )
{
	ChoicePattern simple_cp( new_pattern, D );
	double codelen_as_simple_pattern = simple_cp.get_codelen( D );
	int32_t min_idx = -1;
	double min = 0;
	for(unsigned int cp_idx = 0; cp_idx < list_of_choice_patterns.size(); cp_idx++ )
	{
		double curr_gain = list_of_choice_patterns[ cp_idx ].change_in_codelen_if_added( new_pattern, D ) - codelen_as_simple_pattern;
		if( curr_gain < 0 && curr_gain < min )
		{
			min = curr_gain;
			min_idx = cp_idx;
		}
	}
	return min_idx;
}

void build_list_of_choice_patterns( vector<Pattern*> &list_of_patterns, vector<ChoicePattern> &list_of_choice_patterns, Data *D )
{
	for(uint32_t pat_idx = 0; pat_idx < list_of_patterns.size(); pat_idx++ )
	{
		int choice_pattern_that_can_be_added = check_if_new_pattern_can_be_made_a_choice_pattern( list_of_patterns[ pat_idx ], list_of_choice_patterns, D );
		if( choice_pattern_that_can_be_added != -1 )
		{
			list_of_choice_patterns[ choice_pattern_that_can_be_added ].add_pattern( list_of_patterns[ pat_idx ], D );
		}
		else
		{
			ChoicePattern simple_cp( list_of_patterns[ pat_idx ], D );
			list_of_choice_patterns.push_back( simple_cp );
		}
	}
}

void print_all_choice_patterns_with_labels( vector<ChoicePattern> &list_of_choice_patterns, Data *D, LabelList &label_list )
{
	cout << " ------------ All choice patterns -------- " << endl;
	for(int i = 0; i < list_of_choice_patterns.size(); i++)
	{
		list_of_choice_patterns[i].print( label_list );
		cout << "codelen: " << list_of_choice_patterns[i].get_codelen(D) << endl;
	}
	cout << " ----------------------------------------- " << endl;
}


