/*
 * create_interleaved_data.cpp
 *
 *  Created on: Apr 11, 2016
 *      Author: abhattac
 */

#include "Pattern.h"
#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <string.h>
#include <stack>

using namespace std;

class Context
{
	Pattern *pattern;
	int char_index = 0;
	int gap_stream_index = 0;
	vector<char> gap_stream;
	int gap_stream_len = 0;

	double probability_of_gap = 0.3;

	void create_gap_stream( )
	{
		int fills = calc_fills();
		for(int i = 0; i < fills - 1; i++ )
		{
			gap_stream.push_back( 'f' );
		}
		for(int i = fills - 1; i < gap_stream_len - 2; i++ )
		{
			gap_stream.push_back( 'g' );
		}
	}

	int calc_fills( )
	{
		return pattern->getSize() - 1;
	}


	int calc_gap_stream_len()
	{
		int fills = calc_fills();
		int gaps = (int)(double(fills) * probability_of_gap);
		return fills + gaps ;
	}

public:
	Context( Pattern *p )
	{
		this->pattern = p;
		int gap_stream_len = calc_gap_stream_len();
		this->gap_stream_len = calc_gap_stream_len();
		create_gap_stream();
		random_shuffle( gap_stream.begin(), gap_stream.end() );
		gap_stream.push_back( 'f' );
	}

	int get_next_char()
	{
		if( char_index < pattern->getSize() )
		{
			return pattern->charAt( char_index++ );
		}
		else
		{
			return -1;
		}
	}

	char get_next_gap_stream_char()
	{
		if( gap_stream_index < gap_stream_len )
		{
			return gap_stream[ gap_stream_index++ ];
		}
		else
		{
			return -1;
		}
	}

	bool empty()
	{
		return char_index >= pattern->getSize() ;
	}

};

void create_dummy_patterns( vector<Pattern*> &p_list )
{
	for( int start = 0; start < 10; start++ )
	{
		int tmp_pat[5] = { 5*start, 5*start + 1, 5*start + 2, 5*start + 3, 5*start + 4 };
		vector<int> tmp_pat_vec ( tmp_pat, tmp_pat + sizeof(tmp_pat) / sizeof(int) );
		Pattern *tp = new Pattern( tmp_pat_vec );
		p_list.push_back( tp );
		//tp->printPattern();
	}
}

int calc_pattern_stream_length( vector<Pattern*> &pattern_list, int pattern_freq, int data_len )
{
	for(unsigned int i = 0; i < pattern_list.size(); i++ )
	{
		data_len -= pattern_freq * ( pattern_list[i]->getSize() - 1 );
	}
	return data_len;
}

void account_for_feq_in_complex_patterns( vector<Pattern*> &pattern_list, vector<int> &freq_per_alpha, int pattern_freq )
{
	for(unsigned int i = 0; i < pattern_list.size(); i++ )
	{
		for( int j = 0; j < pattern_list[i]->getSize(); j++ )
		{
			if( freq_per_alpha[ pattern_list[i]->charAt(j) ] >= pattern_freq )
			{
				freq_per_alpha[ pattern_list[i]->charAt(j) ] -= pattern_freq;
			}
		}
	}
}

void create_pattern_stream( vector<int> &pattern_stream, vector<Pattern*> &pattern_list, int num_of_alpha, vector<int> &freq_per_alpha, int pattern_freq )
{
	int alpha_index = 0;
	for(unsigned int i = 0; i < pattern_list.size(); i++ )
	{
		for( int j = 0; j < pattern_freq; j++ )
		{
			pattern_stream.push_back( num_of_alpha + i );
		}
	}
	for(int i = 0; i < num_of_alpha; i++)
	{
		//cout << "Freq per alpha: " << alpha_index << " :" << freq_per_alpha[ alpha_index ] << endl;
		for( int j = 0; j < freq_per_alpha[ alpha_index ]; j++ )
		{
			pattern_stream.push_back( alpha_index );
		}
		alpha_index++;
	}
}




void create_data( vector<int> &data_stream, vector<Pattern*> &pattern_list, vector<int> &pattern_stream, int num_of_alpha )
{
	unsigned int pattern_stream_index = 0;
	vector<Context> context;
	int context_itr = -1;
	while( pattern_stream_index < pattern_stream.size() )
	{
		cout << "Pattern stream: " << pattern_stream_index << " : " << pattern_stream[ pattern_stream_index ] << endl;
		cout << "Context size: " << context.size() << endl;
		cout << "Context itr: " << context_itr << endl;
 		if( context_itr == -1 && pattern_stream[ pattern_stream_index ] < num_of_alpha )
		{
			data_stream.push_back( pattern_stream[ pattern_stream_index++ ] );
			if( context.size() > 0 )
			{
				context_itr = context.size() - 1;
			}
			continue;
		}

		else if( context_itr == -1 && pattern_stream[ pattern_stream_index ] >= num_of_alpha )
		{
			int pattern_index = pattern_stream[ pattern_stream_index++ ] - num_of_alpha;
			Pattern *p = pattern_list[ pattern_index ];
			Context context_tmp( p );
			context.push_back( context_tmp );
			data_stream.push_back(  context.back().get_next_char() );

			if( context.size() > 0 )
			{
				context_itr = context.size() - 1;
			}
			continue;
		}
		else if( context_itr >=0 )
		{
			char gap_stream_char = context[ context_itr ].get_next_gap_stream_char();
			cout << "Inside context" << endl;
			cout << "Current gap stream char: " << gap_stream_char << endl;
			if( gap_stream_char == 'f' )
			{
				int next_char = context[ context_itr ].get_next_char();
				data_stream.push_back( next_char );
				if( context[ context_itr ].empty() )
				{
					Context old_context = context[ context_itr ];
					context.erase( context.begin() + context_itr );
				}
				if( context.size() > 0 )
				{
					context_itr = context.size() - 1;
				}
				else
				{
					context_itr = -1;
				}
				continue;
			}
			else if( gap_stream_char == 'g' )
			{
				context_itr--;
			}
			/*else
			{
				return;
			}*/
		}
	}
}

void write_data_stream_to_file( vector<int> &data_stream, char *filename )
{
	ofstream data_stream_to_file;
	data_stream_to_file.open( filename );
	for(unsigned int  i = 0; i < data_stream.size(); i++ )
	{
		//cout << data_stream[i] << "," << endl;
		data_stream_to_file << data_stream[i] << ",";
	}
	data_stream_to_file.close();
}

void create_interleaved_data( int num_of_alpha, int data_len, int pattern_freq, char *filename )
{
	vector<Pattern*> pattern_list;
	create_dummy_patterns( pattern_list );
	cout << "Dummy patterns created" << endl;

	vector<int> freq_per_alpha ( num_of_alpha, data_len / num_of_alpha);
	account_for_feq_in_complex_patterns( pattern_list, freq_per_alpha, pattern_freq );


	vector<int> pattern_stream;
	create_pattern_stream( pattern_stream, pattern_list, num_of_alpha, freq_per_alpha, pattern_freq );

	random_shuffle( pattern_stream.begin(), pattern_stream.end() );
	cout << "Pattern stream created" << endl;

	vector<int> data_stream;

	create_data( data_stream, pattern_list, pattern_stream, num_of_alpha );
	write_data_stream_to_file( data_stream, filename );
}



