/*
 * PostProcess.cpp
 *
 *  Created on: Jun 28, 2016
 *      Author: abhattac
 */
#include <cstdint>
#include <vector>
#include "Pattern.h"

typedef uint32_t uint32;

uint32 min(uint32 x, uint32 y, uint32 z)
{
    return min(min(x, y), z);
}

uint32 edit_dist_DP( Pattern* p1, Pattern* p2 )
{
    // Create a table to store results of subproblems
	uint32 m = p1->getSize(), n = p2->getSize();
    uint32 dp[m+1][n+1];

    // Fill d[][] in bottom up manner
    for (uint32 i = 0; i <= m; i++ )
    {
        for (uint32 j = 0; j <= n; j++ )
        {
            // If first string is empty, only option is to
            // isnert all characters of second string
            if( i == 0 )
                dp[i][j] = j;  // Min. operations = j

            // If second string is empty, only option is to
            // remove all characters of second string
            else if (j==0)
                dp[i][j] = i; // Min. operations = i

            // If last characters are same, ignore last char
            // and recur for remaining string
            else if (p1->charAt(i - 1) == p2->charAt(j - 1))
                dp[i][j] = dp[i-1][j-1];

            // If last character are different, consider all
            // possibilities and find minimum
            else
                dp[i][j] = 1 + min(dp[i][j-1],  // Insert
                                   dp[i - 1][j],  // Remove
                                   dp[i - 1][j-1]); // Replace
        }
    }

    return dp[m][n];
}

bool check_if_pattern_is_a_subsequence( Pattern* p1, Pattern* p2 )
{
	int l1 = p1->getSize();
	int l2 = p2->getSize();
	int l3 = l1-l2; // avoid repeated computations
	for(int i = 0;i<l3;i++){
	    bool matched = true; // assume substring match until first mismatch
	    for(int j1 = i, j2 = 0; j2 < l2; j1++, j2++) {
	        if ( p1->charAt(j1) != p2->charAt(j2) ) {
	            matched = false;
	            break; // early out of inner loop as soon as substring mismatch
	        }
	    }
	    if (matched) return true; // no mismatches means found 1st substring
	}
	return false; // nothing found here means no substring at all
}

double avg_isd( vector<Pattern*> &list_of_patterns )
{
	double isd = 0;
	uint32 count = 0;
	for(uint32 i = 0; i < list_of_patterns.size(); i++ )
	{
		if( list_of_patterns[i]->getSortIndex() < 50 )
		{
			for(uint32 j = 0; j < list_of_patterns.size(); j++ )
			{
				if( list_of_patterns[j]->getSortIndex() < 50 )
				{
					isd += edit_dist_DP( list_of_patterns[i], list_of_patterns[j] );
					count++;
				}
			}
		}
	}
	return isd / (count);
}

double avg_cs( vector<Pattern*> &list_of_patterns )
{
	double cs = 0, count = 0;
	for(uint32 i = 0; i < list_of_patterns.size(); i++ )
	{
		if( list_of_patterns[i]->getSortIndex() < 50 )
		{
			for(uint32 j = 0; j < list_of_patterns.size(); j++ )
			{
				if(  list_of_patterns[j]->getSortIndex() < 50 && check_if_pattern_is_a_subsequence( list_of_patterns[i], list_of_patterns[j] ) )
				{
					cs ++;
					break;
				}
			}
			count++;
		}
	}
	return cs / (count);
}
